@@ -1157,6 +1157,12 @@ ansi_send_home()
 	net_sends("\033[H");
 }
 
+void
+ansi_send_clear()
+{
+	net_sends("\033[2K");
+}
+
 void
 ansi_send_pf(nn)
 int nn;
