@@ -133,14 +133,8 @@ char *mn;
 		break;
 	default:
 	case 4:
-		if (atoi(mn) != 4) {
-			char	*buf;
-
-			buf = xs_buffer("Unknown model %s, defaulting to 4",
-			    mn);
-			XtWarning(buf);
-			XtFree(buf);
-		}
+		if (atoi(mn) != 4)
+			xs_warning("Unknown model %s, defaulting to 4", mn);
 		maxROWS = ROWS = 43; 
 		maxCOLS = COLS = 80;
 		model_num = 4;
@@ -1022,9 +1016,9 @@ int	move_ea;
 	if (memcmp((char *) &screen_buf[baddr_from],
 	           (char *) &screen_buf[baddr_to],
 		   count)) {
-		(void) memcpy((char *) &screen_buf[baddr_to],
-			      (char *) &screen_buf[baddr_from],
-			      count);
+		(void) MEMORY_MOVE((char *) &screen_buf[baddr_to],
+			           (char *) &screen_buf[baddr_from],
+			           count);
 		screen_changed = True;
 		/*
 		 * For the time being, if any selected text shifts around on
@@ -1038,9 +1032,9 @@ int	move_ea;
 	if (move_ea && memcmp((char *) &ea_buf[baddr_from],
 			      (char *) &ea_buf[baddr_to],
 			      count)) {
-		(void) memcpy((char *) &ea_buf[baddr_to],
-		              (char *) &ea_buf[baddr_from],
-			      count);
+		(void) MEMORY_MOVE((char *) &ea_buf[baddr_to],
+		                   (char *) &ea_buf[baddr_from],
+			           count);
 		screen_changed = True;
 	}
 }
