@@ -26,6 +26,14 @@
 /* Externals: tables.c */
 extern unsigned char	asc2cg[];
 
+/* Statics */
+#define NP	5
+static Atom	paste_atom[NP];
+static int	n_pasting = 0;
+static int	pix = 0;
+static Time	paste_time;
+static enum	{ NONE, COMPOSE, FIRST } composing = NONE;
+
 /* Globals */
 Boolean		kybdlock = True,	/* kybd locked */
 		insert = False;		/* insert mode */
@@ -37,6 +45,23 @@ extern unsigned char	*screen_buf;
 extern int		cursor_addr, buffer_addr;
 extern Boolean		formatted;
 
+/* Composite key mappings. */
+
+enum keytype { LATIN, APL };
+struct akeysym {
+	char keysym;
+	enum keytype keytype;
+};
+static struct akeysym cc_first;
+static struct composite {
+	struct akeysym k1, k2;
+	unsigned char translation;
+} *composites = NULL;
+static int n_composites = 0;
+
+#define ak_eq(k1, k2)	(((k1).keysym  == (k2).keysym) && \
+			 ((k1).keytype == (k2).keytype))
+
 
 /*
  * Called when a host connects or disconnects.
@@ -187,21 +212,22 @@ int	cgcode;
 				ctlr_add(baddr, (unsigned char) cgcode);
 			}
 
-			/*
-			 * If data input moves the cursor onto a field
-			 * attribute, jump over it.  If that attribute
-			 * specified autoskip, skip the field as well.
-			 */
-			INC_BA(baddr);
-			while (IS_FA(screen_buf[baddr])) {
-				Boolean skip = FA_IS_SKIP(screen_buf[baddr]);
 
+			/* Implement auto-skip. */
+			INC_BA(baddr);
+			if (IS_FA(screen_buf[baddr]) &&
+			    FA_IS_SKIP(screen_buf[baddr])) {
+				do {
+					INC_BA(baddr);
+				} while (!IS_FA(screen_buf[baddr]) ||
+					 FA_IS_PROTECTED(screen_buf[baddr]));
 				INC_BA(baddr);
-				if (skip)
-					while (!IS_FA(screen_buf[baddr]))
-						INC_BA(baddr);
 			}
 
+			/* Skip field attribute bytes. */
+			else while (IS_FA(screen_buf[baddr]))
+				INC_BA(baddr);
+
 			cursor_move(baddr);
 
 			mdt_set(fa);
@@ -214,11 +240,55 @@ int	cgcode;
  * Handle an ordinary character key, given an ASCII code.
  */
 static void
-key_ACharacter(c)
+key_ACharacter(c, keytype)
 unsigned char	c;
+enum keytype	keytype;
 {
+	register int i;
+	struct akeysym ak;
+
+	ak.keysym = c;
+	ak.keytype = keytype;
+
+	switch (composing) {
+	    case NONE:
+		break;
+	    case COMPOSE:
+		for (i = 0; i < n_composites; i++)
+			if (ak_eq(composites[i].k1, ak) ||
+			    ak_eq(composites[i].k2, ak))
+				break;
+		if (i < n_composites) {
+			cc_first.keysym = c;
+			cc_first.keytype = keytype;
+			composing = FIRST;
+			status_compose(True, c);
+		} else {
+			ring_bell();
+			composing = NONE;
+			status_compose(False, 0);
+		}
+		return;
+	    case FIRST:
+		composing = NONE;
+		status_compose(False, 0);
+		for (i = 0; i < n_composites; i++)
+			if ((ak_eq(composites[i].k1, cc_first) &&
+			     ak_eq(composites[i].k2, ak)) ||
+			    (ak_eq(composites[i].k1, ak) &&
+			     ak_eq(composites[i].k2, cc_first)))
+				break;
+		if (i < n_composites)
+			c = composites[i].translation;
+		else {
+			ring_bell();
+			return;
+		}
+		break;
+	}
+
 	if (IN_3270) {
-		if (c == '^') {
+		if (c == '^' && keytype == LATIN) {
 			(void) key_Character(CG_NOT);
 		} else {
 			(void) key_Character((int) asc2cg[c]);
@@ -322,6 +392,8 @@ key_Reset()
 	oia_twait = False;
 	oia_locked = False;
 	aid2 = AID_NO;
+	composing = NONE;
+	status_compose(False, 0);
 }
 
 
@@ -332,7 +404,6 @@ void
 key_Home()
 {
 	register int	baddr, nbaddr;
-	register unsigned char	*fa;
 
 	if (IN_ANSI) {
 		ansi_send_home();
@@ -340,26 +411,24 @@ key_Home()
 	}
 	if (kybdlock)
 		return;
-	fa = get_field_attribute((ROWS*COLS)-1);
-	if (!FA_IS_PROTECTED(*fa))
+	if (!formatted) {
 		cursor_move(0);
-	else {
-		nbaddr = 1;		/* start at 2nd col, 1st col is fa */
-		while (True) {
-			baddr = nbaddr;
-			INC_BA(nbaddr);
-			if (IS_FA(screen_buf[baddr])
-			    &&  !FA_IS_PROTECTED(screen_buf[baddr])
-			    &&  !IS_FA(screen_buf[nbaddr]))
-				break;
-			if (baddr == 0) {
-				cursor_move(0);
-				return;
-			}
-		}
-		INC_BA(baddr);
-		cursor_move(baddr);
+		return;
 	}
+	baddr = nbaddr = 0;
+	while (baddr) {
+		INC_BA(nbaddr);
+		if (IS_FA(screen_buf[baddr]) &&
+		    !FA_IS_PROTECTED(screen_buf[baddr]) &&
+		    !IS_FA(screen_buf[nbaddr])) {
+			cursor_move(nbaddr);
+			return;
+		}
+		baddr = nbaddr;
+	}
+
+	/* No unprotected fields */
+	cursor_move(0);
 }
 
 
@@ -390,7 +459,7 @@ void
 key_BackSpace()
 {
 	if (IN_ANSI) {
-		net_sendc('\b');
+		net_send_erase();
 	} else {
 		key_Left();
 	}
@@ -615,7 +684,11 @@ void key_SysReq() { if (!kybdlock) key_AID(AID_SYSREQ); }
 void
 key_Clear()
 {
-	if (IN_ANSI || (kybdlock && CONNECTED))
+	if (IN_ANSI) {
+		ansi_send_clear();
+		return;
+	}
+	if (kybdlock && CONNECTED)
 		return;
 	buffer_addr = 0;
 	ctlr_clear();
@@ -784,7 +857,7 @@ key_Delete()
 		DEC_BA(end_baddr);
 		if (end_baddr > baddr) {
 			ctlr_bcopy(baddr+1, baddr, end_baddr - baddr, 0);
-		} else {
+		} else if (end_baddr != baddr) {
 			ctlr_bcopy(baddr+1, baddr, ((ROWS * COLS) - 1) - baddr, 0);
 			ctlr_add((ROWS * COLS) - 1, screen_buf[0]);
 			ctlr_bcopy(1, 0, end_baddr, 0);
@@ -794,6 +867,142 @@ key_Delete()
 	}
 }
 
+
+/*
+ * Delete word key.  Backspaces the cursor until it hits the front of a word,
+ * deletes characters until it hits a blank or null, and deletes all of these
+ * but the last.
+ *
+ * Which is to say, does a ^W.
+ */
+void
+key_DeleteWord()
+{
+	register int	baddr, baddr2, front_baddr, back_baddr, end_baddr;
+	register unsigned char	*fa;
+
+	if (IN_ANSI) {
+		net_send_werase();
+		return;
+	}
+	if (kybdlock || !formatted)
+		return;
+
+	baddr = cursor_addr;
+	fa = get_field_attribute(baddr);
+
+	/* Make sure we're on a modifiable field. */
+	if (FA_IS_PROTECTED(*fa) || IS_FA(screen_buf[baddr])) {
+		status_protected();
+		mcursor_locked();
+		kybdlock = True;
+		return;
+	}
+
+	/* Search backwards for a non-blank character. */
+	front_baddr = baddr;
+	while (screen_buf[front_baddr] == CG_BLANK ||
+	       screen_buf[front_baddr] == CG_NULLBLANK)
+		DEC_BA(front_baddr);
+
+	/* If we ran into the edge of the field without seeing any non-blanks,
+	   there isn't any word to delete; just move the cursor. */
+	if (IS_FA(screen_buf[front_baddr])) {
+		cursor_move(front_baddr+1);
+		return;
+	}
+
+	/* front_baddr is now pointing at a non-blank character.  Now search
+	   for the first blank to the left of that (or the edge of the field),
+	   leaving front_baddr pointing at the the beginning of the word. */
+	while (!IS_FA(screen_buf[front_baddr]) &&
+	       screen_buf[front_baddr] != CG_BLANK &&
+	       screen_buf[front_baddr] != CG_NULLBLANK)
+		DEC_BA(front_baddr);
+	INC_BA(front_baddr);
+
+	/* Find the end of the word, searching forward for the edge of the
+	   field or a non-blank. */
+	back_baddr = front_baddr;
+	while (!IS_FA(screen_buf[back_baddr]) &&
+	       screen_buf[back_baddr] != CG_BLANK &&
+	       screen_buf[back_baddr] != CG_NULLBLANK)
+		INC_BA(back_baddr);
+
+	/* Find the start of the next word, leaving back_baddr pointing at it
+	   or at the end of the field. */
+	while (screen_buf[back_baddr] == CG_BLANK ||
+	       screen_buf[back_baddr] == CG_NULLBLANK)
+		INC_BA(back_baddr);
+
+	/* Find the end of the field, leaving end_baddr pointing at the field
+	   attribute of the start of the next field. */
+	end_baddr = back_baddr;
+	while (!IS_FA(screen_buf[end_baddr]))
+		INC_BA(end_baddr);
+
+	/* Copy any text to the right of the word we are deleting. */
+	baddr = front_baddr;
+	baddr2 = back_baddr;
+	while (baddr2 != end_baddr) {
+		ctlr_add(baddr, screen_buf[baddr2]);
+		INC_BA(baddr);
+		INC_BA(baddr2);
+	}
+
+	/* Insert nulls to pad out the end of the field. */
+	while (baddr != end_baddr) {
+		ctlr_add(baddr, CG_NULLBLANK);
+		INC_BA(baddr);
+	}
+
+	/* Set the MDT and move the cursor. */
+	mdt_set(fa);
+	cursor_move(front_baddr);
+}
+
+
+
+/*
+ * Delete field key.  Similar to EraseEOF, but it wipes out the entire field
+ * rather than just to the right of the cursor, and it leaves the cursor at
+ * the front of the field.
+ *
+ * Which is to say, does a ^U.
+ */
+void
+key_DeleteField()
+{
+	register int	baddr;
+	register unsigned char	*fa;
+
+	if (IN_ANSI) {
+		net_send_kill();
+		return;
+	}
+	if (kybdlock || !formatted)
+		return;
+
+	baddr = cursor_addr;
+	fa = get_field_attribute(baddr);
+	if (FA_IS_PROTECTED(*fa) || IS_FA(screen_buf[baddr])) {
+		status_protected();
+		mcursor_locked();
+		kybdlock = True;
+		return;
+	}
+	while (!IS_FA(screen_buf[baddr]))
+		DEC_BA(baddr);
+	INC_BA(baddr);
+	cursor_move(baddr);
+	while (!IS_FA(screen_buf[baddr])) {
+		ctlr_add(baddr, CG_NULLBLANK);
+		INC_BA(baddr);
+	}
+	mdt_set(fa);
+}
+
+
 
 /*
  * Set insert mode key.
@@ -813,27 +1022,38 @@ key_Insert()
 
 /*
  * Translate a keymap (from an XQueryKeymap or a KeymapNotify event) into
- * what looks like the state field in a KeyPress event.
+ * a bitmap of Shift, Meta or Alt keys pressed.
  */
-#define key_is_down(kc, bitmap) ((bitmap)[(kc)/8] & (1<<((kc)%8)))
+#define key_is_down(kc, bitmap) (kc && ((bitmap)[(kc)/8] & (1<<((kc)%8))))
 int
 state_from_keymap(keymap)
 char keymap[32];
 {
 	static Boolean	initted = False;
 	static KeyCode	kc_Shift_L, kc_Shift_R;
-	int	fake_state = 0;
+	static KeyCode	kc_Meta_L, kc_Meta_R;
+	static KeyCode	kc_Alt_L, kc_Alt_R;
+	int	pseudo_state = 0;
 
 	if (!initted) {
 		kc_Shift_L = XKeysymToKeycode(display, XK_Shift_L);
 		kc_Shift_R = XKeysymToKeycode(display, XK_Shift_R);
+		kc_Meta_L  = XKeysymToKeycode(display, XK_Meta_L);
+		kc_Meta_R  = XKeysymToKeycode(display, XK_Meta_R);
+		kc_Alt_L   = XKeysymToKeycode(display, XK_Alt_L);
+		kc_Alt_R   = XKeysymToKeycode(display, XK_Alt_R);
 		initted = True;
 	}
 	if (key_is_down(kc_Shift_L, keymap) ||
-	    key_is_down(kc_Shift_R, keymap)) {
-		fake_state |= ShiftMask;
-	}
-	return fake_state;
+	    key_is_down(kc_Shift_R, keymap))
+		pseudo_state |= ShiftKeyDown;
+	if (key_is_down(kc_Meta_L, keymap) ||
+	    key_is_down(kc_Meta_R, keymap))
+		pseudo_state |= MetaKeyDown;
+	if (key_is_down(kc_Alt_L, keymap) ||
+	    key_is_down(kc_Alt_R, keymap))
+		pseudo_state |= AltKeyDown;
+	return pseudo_state;
 }
 #undef key_is_down
 
@@ -861,6 +1081,115 @@ Cardinal *num_params;
 	shift_event(state_from_keymap(keys));
 }
 
+/* Translate a keysym name to a keysym, including APL characters. */
+static KeySym
+StringToKeysym(s, keytypep)
+char *s;
+enum keytype *keytypep;
+{
+	KeySym k;
+
+	if (!strncmp(s, "apl_", 4)) {
+		k = APLStringToKeysym(s);
+		*keytypep = APL;
+	} else {
+		k = XStringToKeysym(s);
+		*keytypep = LATIN;
+	}
+	if (k == NoSymbol && strlen(s) == 1)
+		k = s[0] & 0xff;
+	if (k < ' ' || k > 0xff)
+		k = NoSymbol;
+	return k;
+}
+
+static Boolean
+build_composites()
+{
+	char *c;
+	char *cn;
+	char *ln;
+	char ksname[3][64];
+	char junk[2];
+	KeySym k[3];
+	enum keytype a[3];
+	int i;
+	struct composite *cp;
+
+	if (!appres.compose_map) {
+		XtWarning("Compose: No composeMap defined");
+		return False;
+	}
+	cn = xs_buffer("composeMap.%s", appres.compose_map);
+	if ((c = get_resource(cn)) == NULL) {
+		xs_warning("Compose: Can't find composeMap %s",
+		    appres.compose_map);
+		return False;
+	}
+	XtFree(cn);
+	while (ln = strtok(c, "\n")) {
+		Boolean okay = True;
+
+		c = NULL;
+		if (sscanf(ln, " %63[^+ \t] + %63[^= \t] =%63s%1s",
+		    ksname[0], ksname[1], ksname[2], junk) != 3) {
+			xs_warning("Compose: Invalid syntax: %s", ln);
+			continue;
+		}
+		for (i = 0; i < 3; i++) {
+			k[i] = StringToKeysym(ksname[i], &a[i]);
+			if (k[i] == NoSymbol) {
+				xs_warning("Compose: Invalid KeySym: '%s'",
+				    ksname[i]);
+				okay = False;
+				break;
+			}
+		}
+		if (!okay)
+			continue;
+		composites = (struct composite *) XtRealloc((char *)composites,
+		    (n_composites + 1) * sizeof(struct composite));
+		cp = composites + n_composites;
+		cp->k1.keysym = k[0];
+		cp->k1.keytype = a[0];
+		cp->k2.keysym = k[1];
+		cp->k2.keytype = a[1];
+		cp->translation = k[2];
+		n_composites++;
+	}
+	return True;
+}
+
+/*
+ * Called by the toolkit when the "Compose" key is pressed.  "Compose" is
+ * implemented by pressing and releasing three keys: "Compose" and two
+ * data keys.  For example, "Compose" "s" "s" gives the German "ssharp"
+ * character, and "Compose" "C", "," gives a capital "C" with a cedilla
+ * (symbol Ccedilla).
+ *
+ * The mechanism breaks down a little when the user presses "Compose" and
+ * then a non-data key.  Oh well.
+ */
+/*ARGSUSED*/
+static void
+Compose(w, event, params, num_params)
+Widget w;
+XEvent *event;
+String *params;
+Cardinal *num_params;
+{
+	if (!composites && !build_composites())
+		return;
+
+	if (composing != NONE) {
+		composing = NONE;
+		status_compose(False, 0);
+		return;
+	}
+	composing = COMPOSE;
+	status_compose(True, 0);
+}
+
 /*
  * Called by the toolkit for any key without special actions.
  */
@@ -881,9 +1210,10 @@ Cardinal *num_params;
 	if (ll != 1)
 		return;
 
-	key_ACharacter((unsigned char) buf[0]);
+	key_ACharacter((unsigned char) buf[0], LATIN);
 }
 
+
 /*
  * Key action.
  */
@@ -897,22 +1227,17 @@ Cardinal *nparams;
 {
 	int i;
 	KeySym k;
+	enum keytype keytype;
 
 	for (i = 0; i < *nparams; i++) {
-		k = XStringToKeysym(params[i]);
+		char *s = params[i];
+
+		k = StringToKeysym(s, &keytype);
 		if (k == NoSymbol) {
-			if (strlen(params[i]) == 1)
-				k = params[i][0] & 0xff;
-			else {
-				XtWarning("Key: No such symbol");
-				continue;
-			}
-		}
-		if (k < ' ' || k > 0xff) {
-			XtWarning("Key: Illegal value");
+			xs_warning("Key: Nonexistent or invalid KeySym: %s", s);
 			continue;
 		}
-		key_ACharacter((unsigned char) k);
+		key_ACharacter((unsigned char) k, keytype);
 	}
 }
 
@@ -1009,10 +1334,10 @@ Boolean pasting;
 				if (!pasting)
 					state = BACKSLASH;
 				else
-					key_ACharacter((unsigned char) c);
+					key_ACharacter((unsigned char) c, LATIN);
 				break;
 			default:
-				key_ACharacter((unsigned char) c);
+				key_ACharacter((unsigned char) c, LATIN);
 				break;
 			}
 			break;
@@ -1080,7 +1405,7 @@ Boolean pasting;
 				literal = (literal * 8) + (strchr(dxl, c) - dxl);
 				break;
 			} else {
-				key_ACharacter((unsigned char) literal);
+				key_ACharacter((unsigned char) literal, LATIN);
 				state = BASE;
 				continue;
 			}
@@ -1089,7 +1414,7 @@ Boolean pasting;
 				literal = (literal * 16) + (strchr(dxl, tolower(c)) - dxl);
 				break;
 			} else {
-				key_ACharacter((unsigned char) literal);
+				key_ACharacter((unsigned char) literal, LATIN);
 				state = BASE;
 				continue;
 			}
@@ -1119,16 +1444,19 @@ int *format;
 	unsigned long len;
 
 	if ((value == NULL) || (*length == 0)) {
-#ifdef notdef
-		XtWarning("No selection or selection timed out");
-#endif
 		XtFree(value);
+
+		/* Try the next one. */
+		if (n_pasting > pix)
+			XtGetSelectionValue(w, paste_atom[pix++], XA_STRING,
+			    paste_callback, NULL, paste_time);
 		return;
 	}
 
 	s = (char *)value;
 	len = *length;
 	(void) emulate_input(s, (int) len, True);
+	n_pasting = 0;
 
 	XtFree(value);
 }
@@ -1140,17 +1468,24 @@ XButtonEvent *event;
 String *params;
 Cardinal *nparams;
 {
-	int		i;
+	int	i;
 	Atom	a;
 
+	n_pasting = 0;
 	for (i = 0; i < *nparams; i++) {
 		a = XInternAtom(display, params[i], True);
 		if (a == None) {
 			XtWarning("No atom for selection");
 			continue;
 		}
-		XtGetSelectionValue(w, a, XA_STRING, paste_callback, event,
-		    event->time);
+		if (n_pasting < NP)
+			paste_atom[n_pasting++] = a;
+	}
+	pix = 0;
+	if (n_pasting > pix) {
+		paste_time = event->time;
+		XtGetSelectionValue(w, paste_atom[pix++], XA_STRING,
+		    paste_callback, NULL, paste_time);
 	}
 }
 
@@ -1170,9 +1505,12 @@ XtActionsRec actions[] = {
 	{ "BackSpace",	(XtActionProc)key_BackSpace },
 	{ "BackTab",	(XtActionProc)key_BTab },
 	{ "Clear",	(XtActionProc)key_Clear },
+	{ "Compose",	(XtActionProc)Compose },
 	{ "CursorSelect", (XtActionProc)key_CursorSelect },
 	{ "Default",	(XtActionProc)Default },
 	{ "Delete", 	(XtActionProc)key_Delete },
+	{ "DeleteField",(XtActionProc)key_DeleteField },
+	{ "DeleteWord",	(XtActionProc)key_DeleteWord },
 	{ "DeleteWindow", (XtActionProc)delete_window },
 	{ "Down",	(XtActionProc)key_Down },
 	{ "Dup",	(XtActionProc)key_Dup },
@@ -1190,6 +1528,7 @@ XtActionsRec actions[] = {
 	{ "Left",	(XtActionProc)key_Left },
 	{ "Left2", 	(XtActionProc)key_Left2 },
 	{ "MonoCase",	(XtActionProc)key_MonoCase },
+	{ "MoveCursor",	(XtActionProc)MoveCursor },
 	{ "Newline",	(XtActionProc)key_Newline },
 	{ "PA1",	(XtActionProc)key_PA1 },
 	{ "PA2",	(XtActionProc)key_PA2 },
