@@ -830,10 +830,7 @@ Cardinal *num_params;
 		return;
 	}
 	if (!(c = XtNameToWidget(menu_bar, params[0]))) {
-		char *s = xs_buffer("HandleMenu: can't find menu %s", params[0]);
-
-		XtWarning(s);
-		XtFree(s);
+		xs_warning("HandleMenu: can't find menu %s", params[0]);
 		return;
 	}
 	XtCallActionProc(menu_bar, "XawPositionSimpleMenu", event, params, 1);
@@ -860,11 +857,7 @@ hostfile_init()
 
 	hf = fopen(appres.hostsfile, "r");
 	if (!hf) {
-		char *buf = xs_buffer("Can't read hostsFile '%s'",
-		    appres.hostsfile);
-
-		XtWarning(buf);
-		XtFree(buf);
+		xs_warning("Can't read hostsFile '%s'", appres.hostsfile);
 		return;
 	}
 
