@@ -463,6 +463,7 @@ extern int linemode;
 extern char *build;
 extern char ttype_val[];
 extern Pixmap icon;
+extern Boolean *overstrike;
 
 /* Called when OK is pressed on the options popup */
 /*ARGSUSED*/
@@ -629,11 +630,12 @@ popup_options()
 	XtFree(xbuf);
 
 	if (appres.mono) {
-		if (bfontinfo) {
+		if (*overstrike) {
+			MAKE_LABEL(get_message("overstriking"), 4);
+		} else {
 			MAKE_LABEL(get_message("boldEmulatorFont"), 4);
 			MAKE_VALUE(bfontname);
-		} else
-			MAKE_LABEL(get_message("overstriking"), 4);
+		}
 	}
 
 	if (appres.charset) {
@@ -642,11 +644,17 @@ popup_options()
 	} else
 		MAKE_LABEL(get_message("defaultCharacterSet"), 4);
 
-	if (appres.keymap) {
+	if (appres.key_map) {
 		MAKE_LABEL(get_message("keyboardMap"), 4)
-		MAKE_VALUE(appres.keymap);
+		MAKE_VALUE(appres.key_map);
 	} else
 		MAKE_LABEL(get_message("defaultKeyboardMap"), 4);
+	if (appres.compose_map) {
+		MAKE_LABEL(get_message("composeMap"), 4);
+		MAKE_VALUE(appres.compose_map);
+	} else {
+		MAKE_LABEL(get_message("noComposeMap"), 4);
+	}
 
 	if (appres.active_icon) {
 		MAKE_LABEL(get_message("activeIcon"), 4);
@@ -694,7 +702,7 @@ popup_options()
 			    ns_brcvd, (ns_brcvd == 1) ? get_message("byte") : get_message("bytes"));
 		MAKE_LABEL(fbuf, 4);
 
-		if (IN_ANSI && linemode) {
+		if (IN_ANSI) {
 			struct ctl_char *c = net_linemode_chars();
 			int i;
 
