@@ -79,6 +79,7 @@ static void	toggle_monocase();
 static void	toggle_altcursor();
 static void	toggle_blink();
 static void	toggle_cursorp();
+static void	remap_ascii();
 static void	remap_chars();
 static void	screen_focus();
 static void	make_gc_set();
@@ -124,6 +125,7 @@ static struct sstate *ss = &nss;
 Widget 		*screen = &nss.widget;
 Window		*screen_window = &nss.window;
 unsigned int	**x_image = &nss.image;
+Boolean		*overstrike = &nss.overstrike;
 GC		*gc = nss.gc;
 GC		*invgc = nss.invgc;
 int		*char_width = &nss.char_width;
@@ -203,9 +205,10 @@ Boolean font_changed;
 
 	/* Define graphics contexts and character dimensions. */
 	if (font_changed) {
-		make_gcs(&nss);
 		nss.char_width  = CHAR_WIDTH;
 		nss.char_height = CHAR_HEIGHT;
+		make_gcs(&nss);
+		remap_ascii();
 	}
 
 	/* Initialize the controller. */
@@ -778,8 +781,9 @@ unsigned int *buffer;
 	}
 
 	/* The X display is now correct; update ss->image[]. */
-	(void) memcpy((char *) &ss->image[baddr], (char *) &buffer[baddr],
-	              len*sizeof(unsigned int));
+	(void) MEMORY_MOVE((char *) &ss->image[baddr],
+			   (char *) &buffer[baddr],
+	                   len*sizeof(unsigned int));
 }
 
 /*
@@ -1256,12 +1260,16 @@ void
 shift_event(event_state)
 int event_state;
 {
-	Boolean shifted_now = (event_state & ShiftMask) != 0;
+	static int old_state;
+	Boolean shifted_now = (event_state & ShiftKeyDown) != 0;
 
-	if (shifted_now != shifted) {
-		shifted = shifted_now;
-		status_shift_mode(shifted);
-		keypad_shift();
+	if (event_state != old_state) {
+		old_state = event_state;
+		status_shift_mode(event_state);
+		if (shifted != shifted_now) {
+			shifted = shifted_now;
+			keypad_shift();
+		}
 	}
 }
 
@@ -1404,13 +1412,8 @@ XFontStruct **font;
 		if (XGetFontProperty(f, a_spacing, &svalue)) {
 			if ((Atom) svalue != a_c)
 				return "doesn't use constant spacing";
-		} else {
-			char *buf = xs_buffer("Font %s has no SPACING property",
-			    name);
-
-			XtWarning(buf);
-			XtFree(buf);
-		}
+		} else
+			xs_warning("Font %s has no SPACING property", name);
 	}
 
 	*font = f;
@@ -1555,7 +1558,8 @@ char *s;
 
 /*
  * Parse an EBCDIC character map, a series of pairs of numeric EBCDIC codes
- * and ASCII characters or ISO Latin-1 character names.
+ * and ASCII characters or ISO Latin-1 character names, modifying the
+ * CG-to-EBCDIC maps.
  */
 static void
 remap_chars(spec)
@@ -1585,17 +1589,14 @@ char *spec;
 			s = NULL;
 		}
 		ne++;
-		if (sscanf(line, "%63[^:]: %63s", ebcs, isos) != 2 ||
+		if (sscanf(line, "%63[^: \t]: %63s", ebcs, isos) != 2 ||
 		    !(ebc = strtol(ebcs, (char **)NULL, 0)) ||
 		    !(iso = parse_keysym(isos))) {
 			char	nbuf[16];
-			char	*buf;
 
 			(void) sprintf(nbuf, "%d", ne);
-			buf = xs_buffer2("Can't parse charset '%s', entry %s",
+			xs_warning2("Can't parse charset '%s', entry %s",
 			    appres.charset, nbuf);
-			XtWarning(buf);
-			XtFree(buf);
 			continue;
 		}
 		cg = asc2cg[iso];
@@ -1609,6 +1610,67 @@ char *spec;
 	XtFree(spec);
 }
 
+/*
+ * Parse an ASCII remap table, a series of pairs of keysyms, indicating that
+ * the first ASCII code should be translated into the second, modifying the
+ * ASCII-to-CG map.
+ */
+static void
+remap_ascii()
+{
+	char	*mn;
+	char	*s, *line;
+	char	froms[64], tos[64];
+	char	*nl;
+	unsigned char	from, to, cg;
+	int	ne = 0;
+	extern	unsigned char asc2cg[];
+	static	unsigned char asc2cg_save[256];
+	static	Boolean ever = False;
+
+	/* Restore the original ASCII-to-CG map. */
+	if (!ever) {
+		memcpy((char *) asc2cg_save, (char *) asc2cg, 256);
+		ever = True;
+	} else
+		memcpy((char *) asc2cg, (char *) asc2cg_save, 256);
+
+	/* Find one that matches this font. */
+	mn = xs_buffer("cmap.%s", efontname);
+	s = get_resource(mn);
+	XtFree(mn);
+	if (!s)
+		return;
+
+	/* Parse it. */
+	while (s) {
+		while (isspace(*s))
+			s++;
+		if (!*s)
+			break;
+		line = s;
+		nl = strchr(line, '\n');
+		if (nl) {
+			*nl = '\0';
+			s = nl + 1;
+		} else {
+			s = NULL;
+		}
+		ne++;
+		if (sscanf(line, "%63[^: \t]: %63s", froms, tos) != 2 ||
+		    !(from = parse_keysym(froms)) ||
+		    !(to = parse_keysym(tos))) {
+			char	nbuf[16];
+
+			(void) sprintf(nbuf, "%d", ne);
+			xs_warning2("Can't parse cmap '%s', entry %s",
+			    efontname, nbuf);
+			continue;
+		}
+		asc2cg[from] = asc2cg[to];
+	}
+}
+
 
 /*
  * Initialize the active icon font information.
@@ -1622,10 +1684,8 @@ aicon_font_init()
 	}
 	iss.efontinfo = XLoadQueryFont(display, appres.icon_font);
 	if (iss.efontinfo == NULL) {
-		char *msg = xs_buffer("Can't load iconFont '%s'; activeIcon won't work", appres.icon_font);
-
-		XtWarning(msg);
-		XtFree(msg);
+		xs_warning("Can't load iconFont '%s'; activeIcon won't work",
+		    appres.icon_font);
 		appres.active_icon = False;
 		return;
 	}
@@ -1638,10 +1698,7 @@ aicon_font_init()
 	if (appres.label_icon) {
 		ailabel_font = XLoadQueryFont(display, appres.icon_label_font);
 		if (ailabel_font == NULL) {
-			char *msg = xs_buffer("Can't load iconLabelFont '%s' font; labelIcon won't work", appres.icon_label_font);
-
-			XtWarning(msg);
-			XtFree(msg);
+			xs_warning("Can't load iconLabelFont '%s' font; labelIcon won't work", appres.icon_label_font);
 			appres.label_icon = False;
 			return;
 		}
