@@ -138,6 +138,7 @@ Boolean	*pending;
 {
 	struct servent	*sp;
 	struct hostent	*hp;
+	int			port;
 	struct sockaddr_in	sin;
 	int			i;
 	int			on = 1;
@@ -145,6 +146,18 @@ Boolean	*pending;
 #	define CN_FAIL(t) { popup_an_error(t); return -1; }
 #	define CN_SYSFAIL(t) { popup_an_errno(t, errno); return -1; }
 
+	if (!t_valid) {
+		vintr   = parse_ctlchar(appres.intr);
+		vquit   = parse_ctlchar(appres.quit);
+		verase  = parse_ctlchar(appres.erase);
+		vkill   = parse_ctlchar(appres.kill);
+		veof    = parse_ctlchar(appres.eof);
+		vwerase = parse_ctlchar(appres.werase);
+		vrprnt  = parse_ctlchar(appres.rprnt);
+		vlnext  = parse_ctlchar(appres.lnext);
+		t_valid = 1;
+	}
+
 	*pending = False;
 	n_read = 0;
 
@@ -160,7 +173,9 @@ Boolean	*pending;
 	sbptr = &sbbuf[0];
 
 	/* get the tcp/ip service (telnet) */
-	if ((sp = getservbyname(portname, "tcp")) == (struct servent *) 0)
+	if (sp = getservbyname(portname, "tcp"))
+		port = sp->s_port;
+	else if (!(port = atoi(portname)))
 		CN_FAIL("unknown service");
 
 	/* fill in the socket address of the given host */
@@ -178,11 +193,11 @@ Boolean	*pending;
 	}
 	else {
 		sin.sin_family = hp->h_addrtype;
-		(void) memcpy((char *) &sin.sin_addr,
-			      (char *) hp->h_addr,
-			      hp->h_length);
+		(void) MEMORY_MOVE((char *) &sin.sin_addr,
+			           (char *) hp->h_addr,
+				   hp->h_length);
 	}
-	sin.sin_port = sp->s_port;
+	sin.sin_port = port;
 
 	/* create the socket */
 	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
@@ -949,20 +964,8 @@ int init;
 		if (!init)
 			trace("Operating in %s mode.\n",
 			    linemode ? "line" : "character-at-a-time");
-		if (linemode) {
+		if (linemode)
 			cooked_init();
-			if (!t_valid) {
-				vintr   = parse_ctlchar(appres.intr);
-				vquit   = parse_ctlchar(appres.quit);
-				verase  = parse_ctlchar(appres.erase);
-				vkill   = parse_ctlchar(appres.kill);
-				veof    = parse_ctlchar(appres.eof);
-				vwerase = parse_ctlchar(appres.werase);
-				vrprnt  = parse_ctlchar(appres.rprnt);
-				vlnext  = parse_ctlchar(appres.lnext);
-				t_valid = 1;
-			}
-		}
 	}
 }
 
@@ -1080,6 +1083,39 @@ char	*s;
 	net_cookout(s, strlen(s));
 }
 
+
+/*
+ * net_send_erase
+ *	Sends the KILL character in ANSI mode.
+ */
+void
+net_send_erase()
+{
+	net_cookout(&verase, 1);
+}
+
+
+/*
+ * net_send_kill
+ *	Sends the KILL character in ANSI mode.
+ */
+void
+net_send_kill()
+{
+	net_cookout(&vkill, 1);
+}
+
+
+/*
+ * net_send_werase
+ *	Sends the WERASE character in ANSI mode.
+ */
+void
+net_send_werase()
+{
+	net_cookout(&vwerase, 1);
+}
+
 
 /*
  * External entry points to negotiate line or character mode.
