@@ -15,10 +15,10 @@
 #include <X11/StringDefs.h>
 #include <X11/Core.h>
 #include <X11/Shell.h>
-#include "globals.h"
 #include <string.h>
 #include <stdio.h>
 #include <signal.h>
+#include "globals.h"
 
 #include "x3270.bm"
 
@@ -77,6 +77,7 @@ static void		x_throttle();
 
 static XrmOptionDescRec options[]= {
 	{ "-activeicon",".activeIcon",	XrmoptionNoArg,		"true" },
+	{ "-apl",	".aplMode",	XrmoptionNoArg,		"true" },
 	{ "-bfont",	".boldFont",	XrmoptionSepArg,	NULL },
 	{ "-charset",	".charset",	XrmoptionSepArg,	NULL },
 	{ "-efont",	".emulatorFont",XrmoptionSepArg,	NULL },
@@ -123,16 +124,20 @@ static XtResource resources[] = {
 	  offset(bfontname), XtRString, 0, },
 	{ "visualBell", "VisualBell", XtRBoolean, sizeof(Boolean),
 	  offset(visual_bell), XtRString, "false" },
+	{ "aplMode", "AplMode", XtRBoolean, sizeof(Boolean),
+	  offset(apl_mode), XtRString, "false" },
 	{ "model", "Model", XtRString, sizeof(char *),
 	  offset(model), XtRString, "4", },
 	{ "keymap", "Keymap", XtRString, sizeof(char *),
-	  offset(keymap), XtRString, 0, },
+	  offset(key_map), XtRString, 0, },
+	{ "composeMap", "ComposeMap", XtRString, sizeof(char *),
+	  offset(compose_map), XtRString, "latin1", },
 	{ "hostsFile", "HostsFile", XtRString, sizeof(char *),
 	  offset(hostsfile), XtRString, "/usr/local/pub/ibm_hosts", },
 	{ "port", "Port", XtRString, sizeof(char *),
 	  offset(port), XtRString, "telnet", },
 	{ "charset", "Charset", XtRString, sizeof(char *),
-	  offset(charset), XtRString, 0, },
+	  offset(charset), XtRString, "us", },
 	{ "termName", "TermName", XtRString, sizeof(char *),
 	  offset(termname), XtRString, 0, },
 	{ "largeFont", "LargeFont", XtRString, sizeof(char *),
@@ -207,6 +212,7 @@ char	*argv[];
 	Boolean	default_bold = False;
 	char	*ef, *bf = NULL, *emsg;
 	int	i;
+	Boolean saw_apl_keymod = False;
 
 	/* Figure out who we are */
 	programname = strrchr(argv[0], '/');
@@ -271,6 +277,11 @@ char	*argv[];
 	 * say anything if it's wrong.  Finally, fall back to "fixed", and
 	 * die if that won't work.
 	 */
+	if (appres.apl_mode) {
+		if ((appres.efontname = get_resource("aplFont")) == NULL)
+			XtWarning("Can't find aplFont resource");
+		appres.bfontname = NULL;
+	}
 	if (!appres.efontname)
 		appres.efontname = appres.large_font;
 	ef = appres.efontname;
@@ -278,10 +289,8 @@ char	*argv[];
 		bf = appres.bfontname;
 	if (emsg = load_fixed_font(ef, efontinfo)) {
 		bf = NULL;
-		buf = xs_buffer2("emulatorFont '%s' %s, using default font",
+		xs_warning2("emulatorFont '%s' %s, using default font",
 		    ef, emsg);
-		XtWarning(buf);
-		XtFree(buf);
 		if (ef = get_resource("font"))
 			(void) load_fixed_font(ef, efontinfo);
 	}
@@ -304,11 +313,8 @@ char	*argv[];
 			default_bold = True;
 		}
 		(void) load_fixed_font(bf, bfontinfo);
-		if (!*bfontinfo && !default_bold) {
-			buf = xs_buffer("Can't load boldFont '%s'", bf);
-			XtWarning(buf);
-			XtFree(buf);
-		}
+		if (!*bfontinfo && !default_bold)
+			xs_warning("Can't load boldFont '%s'", bf);
 	}
 
 	set_font_globals(*efontinfo, ef, bf);
@@ -316,18 +322,20 @@ char	*argv[];
 	set_rows_cols(appres.model);
 	hostfile_init();
 
-	if (appres.keymap == NULL)
-		appres.keymap = (char *)getenv("KEYMAP");
-	if (appres.keymap == NULL)
-		appres.keymap = (char *)getenv("KEYBD");
-	if (appres.keymap != NULL) {
-		char *ns = XtNewString(appres.keymap);
+	if (appres.key_map == NULL)
+		appres.key_map = (char *)getenv("KEYMAP");
+	if (appres.key_map == NULL)
+		appres.key_map = (char *)getenv("KEYBD");
+	if (appres.key_map != NULL) {
+		char *ns = XtNewString(appres.key_map);
 		char *comma;
 
 		do {
 			comma = strchr(ns, ',');
 			if (comma)
 				*comma = '\0';
+			if (!strcmp(ns, "apl"))
+				saw_apl_keymod = True;
 			find_keymaps(ns);
 			if (comma)
 				ns = comma + 1;
@@ -335,16 +343,23 @@ char	*argv[];
 				ns = NULL;
 		} while (ns);
 	}
+	if (appres.apl_mode && !saw_apl_keymod) {
+		find_keymaps("apl");
+		if (appres.key_map)
+			appres.key_map = xs_buffer("%s,apl", appres.key_map);
+		else
+			appres.key_map = "apl";
+	}
+	if (appres.apl_mode) {
+		appres.compose_map = "apl";
+		appres.charset = "apl";
+	}
 	if (appres.charset != NULL) {
 		buf = xs_buffer("charset.%s", appres.charset);
 		charset = get_resource(buf);
 		XtFree(buf);
-		if (charset == NULL) {
-			buf = xs_buffer("Can't find charset '%s'",
-			    appres.charset);
-			XtWarning(buf);
-			XtFree(buf);
-		}
+		if (charset == NULL)
+			xs_warning("Can't find charset '%s'", appres.charset);
 	}
 
 	if (!strcmp(appres.keypad, "right"))
@@ -752,7 +767,7 @@ XEvent *e;
 }
 
 /*
- * Common helper functions to insert a string, through a template, into a new
+ * Common helper functions to insert strings, through a template, into a new
  * buffer.
  * 'format' is assumed to be a printf format string with one '%s' in it.
  */
@@ -780,3 +795,56 @@ char *string2;
 	(void) sprintf(r, format, string1, string2);
 	return r;
 }
+
+/* Common uses of xs_buffer. */
+void
+xs_warning(format, string)
+char *format;
+char *string;
+{
+	char *r;
+
+	r = xs_buffer(format, string);
+	XtWarning(r);
+	XtFree(r);
+}
+
+void
+xs_warning2(format, string1, string2)
+char *format;
+char *string1;
+char *string2;
+{
+	char *r;
+
+	r = xs_buffer2(format, string1, string2);
+	XtWarning(r);
+	XtFree(r);
+}
+
+#ifndef MEMORY_MOVE
+/*
+ * A version of memcpy that handles overlaps
+ */
+char *
+MEMORY_MOVE(dst, src, cnt)
+register char *dst;
+register char *src;
+register int cnt;
+{
+	char *r = dst;
+
+	if (dst < src && dst + cnt - 1 >= src) {	/* overlap left */
+		while (cnt--)
+			*dst++ = *src++;
+	} else if (src < dst && src + cnt - 1 >= dst) {	/* overlap right */
+		dst += cnt;
+		src += cnt;
+		while (cnt--)
+			*--dst = *--src;
+	} else {					/* no overlap */ 
+		(void) memcpy(dst, src, cnt);
+	}
+	return r;
+}
+#endif
